#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "configfile.h"

#define my_isblank(a) (((a) == '\t') || ((a) == ' ') || ((a) == '\r') || ((a) == '\n'))

static inline int hash(char * string)
{
	int i, h = 0;
	for(i = 0; string[i]; i++) h += string[i];
	return(h % CSS_HASHSIZE);
}

#define BUFFER_STEP 1024
char * css_LoadFile(const char * name, int *lenstore, int err)
{
	char lbuffer[BUFFER_STEP];
	char * ret = NULL;
	int fd, len = 0, l;
	
	fd = open(name, O_RDONLY);
	if(fd < 0)
	{
		if(err) perror("open");
		return(NULL);
	}
	
	while((l = read(fd, lbuffer, BUFFER_STEP)) > 0)
	{
		assert(ret = (char*) realloc(ret, len + l + 1));
		memcpy(ret + len, lbuffer, l);
		len += l;
		ret[len] = 0;
	}
	
	if(l < 0)
	{
		if(err)
			perror("read");
		free(ret);
		return(NULL);
	}
	
	close(fd);
	if(lenstore) *lenstore = len;
	return(ret);
}

cssctx * css_Parse(char * data, int len, char *errbuf, int errlen)
{
	cssctx * ctx;
	int error = 0;
	char * newdata;
	
	assert(newdata = (char*) malloc(len + 1));
	memcpy(newdata, data, len);
	newdata[len] = 0;
	data = newdata;
	
	assert(ctx = (cssctx*) malloc(sizeof(cssctx)));
	memset(ctx, 0, sizeof(cssctx));

	
//	printf("starting a parse\n");
	
	while(len > 0)
	{
		char * token_start;
		char keybuffer[CSS_MAXVALUE];
		char valuebuffer[CSS_MAXVALUE];
		int klen = 0, vlen = 0;
		
//		printf("Context: %20.20s\n", data);
		
		while((my_isblank(*data) || (*data == ';')) && (len > 0))
			{ data++; len--; }
		if(!len) break;
		
		if(*data == '#')
		{
			while((*data != '\n') && (len > 0))
				{ data++; len--; }
			if(!len) break;
			continue;
		}
		
		token_start = data;
		while((*data != ':') && (*data != '{') && (len > 0))
		{
			assert(klen < CSS_MAXVALUE);
			keybuffer[klen++] = *data;
			data++;
			len--;
		}
		
		if(!len) goto tooshort;
		if(klen > 0)
		{
			while(my_isblank(keybuffer[klen - 1]) && (klen > 0))
				klen--;
		}
		
		if(!klen)
		{
			if(errbuf)
				snprintf(errbuf, errlen, "Empty key around: (...) %20.20s", data);
				
			error = 1;
			break;
		}
		
		keybuffer[klen] = 0;
		
		if(*data == ':')	/* a value! */
		{
			cssval * v;
			int h;

			data++; len--;			
			while(my_isblank(*data) && (len > 0)) { data++; len--; }
			if(!len) goto tooshort;
			
			/* now, check encoding type */
			if(*data == '\a') /* literal encoding upto a single quote */
			{
				data++; len--; if(!len) goto tooshort;
				while((*data != '\a') && (len > 0))
				{
					assert(vlen < CSS_MAXVALUE);
					valuebuffer[vlen++] = *data;
					data++;
					len--;
				}
				if(!len) goto tooshort;
				while(my_isblank(*data) && (len > 0))
				{
					data++;
					len--;
				}
				if(!len) goto tooshort;
				if(*data != ';')
				{
					if(errbuf)
						snprintf(errbuf, errlen, "Unidentified characters around: (...) %20.20s", data);
				
					error = 1;
					break;
				}
				data++; len--;
			}
			else if(*data == '"') /* literal encoding w/special sequences */
			{
				data++; len--; if(!len) goto tooshort;
				while(len > 0)
				{
					char c;
					c = *data;
					if(c == '"') break;
					if(c == '\\')
					{
						data++; len--;
						if(!len) goto tooshort;
						c = *data;
						switch(c)
						{
							case 'n': c = '\n'; break;
							case 'r': c = '\r'; break;
							case 'a': c = '\a'; break;
							case 'x':
							{
								char b1, b2;
								if(len < 2) goto tooshort;
								b1 = data[1];
								b2 = data[2];
								data += 2;
								len -= 2;
								if(b1 >= 'a' && b1 <= 'f') b1 = b1 - 'a' + 10;
								if(b1 >= 'A' && b1 <= 'F') b1 = b1 - 'A' + 10;
								if(b2 >= 'a' && b2 <= 'f') b2 = b2 - 'a' + 10;
								if(b2 >= 'A' && b2 <= 'F') b2 = b2 - 'A' + 10;
								c = (b1 << 4) | b2;
								break;
							}
						}
					}
					assert(vlen < CSS_MAXVALUE);
					valuebuffer[vlen++] = c;
					data++; len--;
				}
				if(!len) goto tooshort;
				while(my_isblank(*data) && (len > 0))
				{
					data++;
					len--;
				}
				if(!len) goto tooshort;
				if(*data != ';')
				{
					if(errbuf)
						snprintf(errbuf, errlen, "Unidentified characters around: (...) %20.20s", data);
				
					error = 1;
					break;
				}
				data++; len--;
			}
			else /* up to the semicolon at the end */
			{
				while((*data != ';') && (len > 0))
				{
					assert(vlen < CSS_MAXVALUE);
					valuebuffer[vlen++] = *data;
					data++; len--;
				}
				if(!len) goto tooshort;
				data++; len--;
			}
			assert(vlen < CSS_MAXVALUE);
			valuebuffer[vlen++] = 0;
			
			if(strcmp(keybuffer, "include") == 0)
			{
				int inclen;
				char * incl;
				
				incl = css_LoadFile(valuebuffer, &inclen, 1);
				if(!incl)
				{
					if(errbuf)
						snprintf(errbuf, errlen, "Unable to read file requested for inclusion: %s", valuebuffer);
				
					error = 1;
					break;
				}
				
				assert(incl = (char*) realloc(incl, inclen + len + 1));
				memcpy(incl + inclen, data, len);
				len += inclen;
				incl[len] = 0;
				free(newdata);
				newdata = data = incl;
//			printf("after inclusion: %s\n", data);
				continue;
			}


			assert(v = (cssval*) malloc(sizeof(cssval)));
			
			assert(v->name = strdup(keybuffer));
			assert(v->value = strdup(valuebuffer));
			
			h = hash(v->name);
			
			v->next = ctx->values[h];
			ctx->values[h] = v;
		}
		else /* another context */
		{
			/* skip to } */
			int h;
			char * context_start;
			int clen, level = 0;
			cssctx * child;
			
			data++; len--;
			context_start = data;
			clen = len;
			while(len > 0)
			{
				char c = *(data++);
				len--;
				if(c == '{') { level++; continue; }
				if(c == '}')
				{
					level--;
					if(level < 0) break;
					continue;
				}
				if(c == '#')
				{
					while((len > 0) && *data != '\n') { data++; len--; };
					continue;
				}
				if(c == '\a')
				{
					while((len > 0) && *data != '\a') { data++; len--; };
					continue;
				}
				if(c == '"')
				{
					while((len > 0) && *data != '"') { if(*data == '\\') { data++; len--; }; data++; len--; };
					continue;
				}
			}

			if(!len) goto tooshort;
			
			child = css_Parse(context_start, clen - len - 1, errbuf, errlen);
			if(!child)
			{
				error = 1;
				break;
			}
			assert(child->name = strdup(keybuffer));
			h = hash(child->name);
			child->next = ctx->children[h];
			ctx->children[h] = child;
		}
		
	}
//	printf("alright free.\n");
	free(newdata);
//	printf("alright free done.\n");
	goto chkerror;

tooshort:
	if(errbuf)
		snprintf(errbuf, errlen, "Parse error at the end of input.");
	error = 1;

chkerror:
	if(error)
	{
		free(ctx);
		return(NULL);
	}
	
//	printf("alright.\n");

	return(ctx);
}

cssctx * css_ParseFile(const char * file, char *errbuf, int errlen)
{
	int lenstore;
	char * data;
	cssctx * ctx;
	
	data = css_LoadFile(file, &lenstore, 1);

	if(!data)
	{
		snprintf(errbuf, errlen, "Unable to read file: %s", file);
		return(NULL);
	}
	
	ctx = css_Parse(data, lenstore, errbuf, errlen);
	free(data);
	return(ctx);
}

cssval * css_FindValue(cssctx * ctx, cssval * prev, char *name)
{
	int h;
	
	if(prev)
	{
		for(prev = prev->next; prev; prev = prev->next)
			if(strcmp(name, prev->name) == 0) return(prev);
		return(NULL);
	}
	h = hash(name);
	for(prev = ctx->values[h]; prev; prev = prev->next)
		if(strcmp(name, prev->name) == 0) return(prev);
	return(NULL);
}

cssctx * css_FindChild(cssctx * ctx, cssctx * prev, char *name)
{
	int h;
	
	if(prev)
	{
		name = prev->name;
		for(prev = prev->next; prev; prev = prev->next)
			if(strcmp(name, prev->name) == 0) return(prev);
		return(NULL);
	}
	h = hash(name);
	for(prev = ctx->children[h]; prev; prev = prev->next)
		if(strcmp(name, prev->name) == 0) return(prev);
	return(NULL);
}

int css_TrueString(char *str)
{
	if(strcmp(str, "enable") == 0) return(1);
	if(strcmp(str, "enabled") == 0) return(1);
	if(strcmp(str, "yes") == 0) return(1);
	if(strcmp(str, "on") == 0) return(1);
	if(strcmp(str, "1") == 0) return(1);
	return(0);
}

int css_IsTrue(cssctx * ctx, char * name, int def)
{
	cssval * val = css_FindValue(ctx, NULL, name);
	if(!val) return(def);
	return(css_TrueString(val->value));
}

