#include <string.h>
#include <stdio.h>
#include "memory.h"
#include "hash.h"
#include "debug.h"

inline static int hash(char *key)
{
	int q = 0;
	int i;
	for(i = 0; key[i]; i++)
		q += key[i];
	return q;
}

Hash * hash_create(int size)
{
	Hash *h;
	h = (Hash*) mem_alloc("Hash information storage", sizeof(Hash));
	h->hash_size = size;
	h->entries = (struct hash_entry**) mem_alloc("Hash table", sizeof(struct hash_entry*) * size);
	
	return h;
}

void hash_add(Hash *h, char *k, void *p)
{
	int v = hash(k) % h->hash_size;
	struct hash_entry * e;
	
	e = (struct hash_entry*) mem_alloc("Hash entry storage", sizeof(struct hash_entry));
	e->key = mem_strdup("Hash key", k);
	e->ptr = p;
	e->next = h->entries[v];
	h->entries[v] = e;
}

void * hash_get(Hash *h, char *k)
{
	int v = hash(k) % h->hash_size;
	struct hash_entry * e;
	
	debug("  --> Looking for: %s in %p root entry at place %d = %p\n", k, h, v, h->entries[v]);
	
	for(e = h->entries[v]; e; e = e->next)
	{
		debug("       --> Candidate %s\n", e->key);
		if(!strcmp(k, e->key)) return(e->ptr);
	}
	return(NULL);
}

void hash_remove(Hash *h, char *k)
{
	int v = hash(k) % h->hash_size;
	struct hash_entry **storage;
	
	debug("  --> Removing key %s\n", k);
	
	for(storage = &h->entries[v]; *storage; storage = &((*storage)->next))
		if(!strcmp(k, (*storage)->key))
		{
			struct hash_entry *next = (*storage)->next;

			debug("    --> Removing entry %s at %p\n", k, *storage);
			
			mem_free((*storage)->key);
			mem_free(*storage);
			
			*storage = next;
			if(!next)
				break;
		}

	debug("    --> Removing finished\n");

}
