#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <assert.h>
#include <netinet/ip_icmp.h>
#include <sys/time.h>
#include "hash.h"
#include "aprobe.h"
#include "debug.h"

int in_cksum(u_short *p, int n)
{
  register u_short answer;
  register long sum = 0;
  u_short odd_byte = 0;

  while( n > 1 )  { sum += *p++; n -= 2; }

  /* mop up an odd byte, if necessary */
  if( n == 1 ) {
      *(u_char *)(&odd_byte) = *(u_char *)p;
      sum += odd_byte;
  }

  sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum += (sum >> 16);			/* add carry */
  answer = ~sum;			/* ones-complement, truncate*/
  return (answer);
}

void ping(struct context *ctx, struct entry *e)
{
	char *buffer;
	struct icmp *i;
	int len;
	struct pinginfo *pi;

	assert(buffer = (char*) malloc(len = (ICMP_MINLEN + ctx->size + sizeof(struct pinginfo))));
	memset(buffer, 0, len);

	i = (struct icmp*) buffer;
	pi = (struct pinginfo*) (buffer + ICMP_MINLEN);

	i->icmp_type = ICMP_ECHO;
	i->icmp_code = 0;
	i->icmp_cksum = 0;
	i->icmp_seq = htons(e->lastid);
	i->icmp_id = ctx->pid;

	gettimeofday(&pi->tv, NULL);
	memcpy(&e->tv[e->lastid % BACKLOG], &pi->tv, sizeof(struct timeval));
	memcpy(&e->last_sent, &pi->tv, sizeof(struct timeval));

	e->lastid++;
	e->sent++;
	
	pi->id = e->id;
	
	i->icmp_cksum = in_cksum((u_short*) i, len);

	if(sendto(ctx->skt, buffer, len, 0, (struct sockaddr *) &e->ip, sizeof(struct sockaddr_in)) != len)
		perror("sendto");

	debug("[ping] Sent a ping to %s\n", inet_ntoa(e->ip.sin_addr));
	
	free(buffer);
}

long timeval_diff(struct timeval *a, struct timeval *b)
{
	long temp;
	temp = (a->tv_sec - b->tv_sec) * 1000000 + (a->tv_usec - b->tv_usec);
	return temp;
}

void ping_process(struct context * ctx, char * buffer, int len, struct sockaddr_in raddr)
{
#define BSIZE 4096
	int hlen, i, ms;
	struct entry *e;
	
	struct ip * ip;
	struct icmp * icmp;
	struct pinginfo *pi;
	
	struct timeval tv;
	
	struct bucket *b;
	

	gettimeofday(&tv, NULL);
	
	if(len < sizeof(struct ip))
	{
		fprintf(stderr, "IP packet too short!\n");
		return;
	}

	ip = (struct ip*) buffer;
	hlen = ip->ip_hl << 2;
	
	if(len < (hlen + ICMP_MINLEN))
	{
		fprintf(stderr, "Packet from %s: too short!\n", inet_ntoa(raddr.sin_addr));
		return;
	}
	
	icmp = (struct icmp*) (buffer + hlen);

	if(icmp->icmp_id != ctx->pid)
		return;

	if(icmp->icmp_type != ICMP_ECHOREPLY)
	{
		fprintf(stderr, "Packet from %s: ICMP Class %d\n", inet_ntoa(raddr.sin_addr), icmp->icmp_type);
		return;
	}
	
	if(len < (hlen + ICMP_MINLEN + sizeof(struct pinginfo)))
	{
		fprintf(stderr, "Packet from %s: reply is too short!\n" , inet_ntoa(raddr.sin_addr));
		return;
	}
	
	pi = (struct pinginfo*) (buffer + ICMP_MINLEN + hlen);
	
	i = ntohs(icmp->icmp_seq);
/*	h = pi->id;
	if(h >= hosts_count || h < 0)
	{
		fprintf(stderr, "Packet from %s: unknown host ID (%d)!\n" , inet_ntoa(raddr.sin_addr), h);
		return;
	}
*/	
	debug("[ping] Recvd a ping from %s\n", inet_ntoa(raddr.sin_addr));
	e = hash_get(ctx->ip_hash, inet_ntoa(raddr.sin_addr));
	if(!e)
	{
		fprintf(stderr, "Packet from %s: Unknown host!\n", inet_ntoa(raddr.sin_addr));
		return;
	}
/*	if(memcmp(&e->ip.sin_addr, &raddr.sin_addr, sizeof(&e->ip.sin_addr)))
	{
		char bf[200];
		snprintf(bf, 200, "%s", inet_ntoa(e->ip.sin_addr));
		fprintf(stderr, "Packet from %s: claims to be %s!\n" , inet_ntoa(raddr.sin_addr), bf);
		return;
	}*/
	
	if((e->lastid - i) > (BACKLOG - 2))
	{
		fprintf(stderr, "Packet from %s: out of backlog!\n" , inet_ntoa(raddr.sin_addr));
		return;
	}
	
	if(!e->tv[i % BACKLOG].tv_sec)
	{
		fprintf(stderr, "Packet from %s: duplicate!\n" , inet_ntoa(raddr.sin_addr));
		return;
	}
	
	e->tv[i % BACKLOG].tv_sec = 0;
	e->recv++;


	ms = timeval_diff(&tv, &pi->tv);
	if(ctx->verbose)	fprintf(stderr, "Packet from %s: %d us rtt\n" , inet_ntoa(raddr.sin_addr), ms);
	e->accounted ++;
	e->sum += ms;
	
	for(b = e->buckets; b; b = b->next)
		if(ms < (b->parent->threshold * 1000.0))
		{
			b->bucket += 1;
			debug("[ping_process] recharged bucket to %.2f\n", b->bucket);
		}
	
#if 0
	printf("sent = %s, ", otv(&dtv[4]));
	printf("sentat = %s, ", otv(&dtv[5]));
	printf("enter = %s, ", otv(&dtv[0]));
	printf("leave = %s, ", otv(&dtv[1]));
	printf("recv = %s, ", otv(&dtv[2]));
	printf("print = %s, ", otv(&dtv[3]));
	printf("\n");
#endif
	
}

void recv_pings(struct context * ctx)
{
	char buffer[BSIZE + 1];
	struct sockaddr_in raddr;
	int slen;
	int len;

	for(;;)
	{
		slen = sizeof(struct sockaddr_in);
		if((len = recvfrom(ctx->skt, buffer, BSIZE, 0, (struct sockaddr *) &raddr, &slen)) < 0)
		{
			if(errno != EAGAIN)
				perror("recvfrom");
			return;
		}
		ping_process(ctx, buffer, len, raddr);
	}
}

extern struct context * ctx;
void send_pings(int sig)
{
	struct timeval tv;
	double td;

	if(!ctx->current)
		ctx->current = ctx->hosts;

	gettimeofday(&tv, NULL);
	
	while(ctx->current)
	{
		double T0 = 
		td = timeval_diff(&tv, &ctx->current->last_sent);
		debug("[send_pings] host = %s td = %.2f delay = %.2f\n", inet_ntoa(ctx->current->ip.sin_addr), td / 1000000, 1.0 / ctx->current->policy->freq);
		if(((td / 1000000) < (1.0 / ctx->current->policy->freq)) && ctx->current->last_sent.tv_sec)
		{
			if(ctx->current->next)
				debug("[send_pings] next = %s\n", inet_ntoa(ctx->current->next->ip.sin_addr));
			else
				debug("[send_pings] next = NULL\n");
			ctx->current = ctx->current->next;
		}
		else
			break;
	}
	
	
	if(ctx->current)
	{
		struct entry *e = ctx->current;
		struct bucket *b;
		int err = 0;
		/*
			The bucket logic:
				The user has supplied a negative bucket parameter "limit" (# of allowed negative hits)
				
				If there are more than "limit" hits in the given time span, we shall raise an error.

				Therefore we shall load the bucket with a speed so that in the whole span the bucket
				consists of "limit" tokens. Every loss/negative answer should decrease the bucket by 1.
				When the bucket drains, raise an error.
				
				Every sent message decreases the token by 1.
				Every received message increases the token back to 1.
				
		*/
		
		/*
			Let's load the buckets
		*/
		for(b = e->buckets; b; b = b->next)
		{
			debug("[send_pings] bucket = %.2f\n", b->bucket);
			b->bucket += (td / 1000000) * (b->parent->limit / b->parent->span);
			
			debug("[send_pings] charged bucket to %.2f\n", b->bucket);
			
			if(b->bucket > b->parent->limit)
				b->bucket = b->parent->limit;
			b->bucket -= 1;
			debug("[send_pings] discharged bucket to %.2f\n", b->bucket);
			
			if(b->bucket < -1)
			{
				b->bucket = -1;
				if(!err)
				{
					err = 1;
					printf("%s %.2f alert %s\n", e->label, 1 / e->policy->freq, b->parent->name ? b->parent->name : "?");
				}
			}
		}
	
		ping(ctx, e);
	}


}

